// (c) 2003, Ivan Yanikov, dobrokot 'at' gmail.com 


#include "ImageLoad.h"
#include "file.h"
#include "error.h"
#include "GL/glu.h"
#include <cstring>
#pragma hdrstop

void Image::PrepareMem(int width, int height, PixType type)
{
  info.width=width;
  info.height=height;
  info.type=type;
  int size=width*height;
  switch (type) {
    case ALPHA:
    case LUMA:
      info.comp=1;
      break;
    case LUMA_ALPHA:
      info.comp=2;
      break;
    case RGB:
      info.comp=3;
      break;
    case RGBA:
      info.comp=4;
      break;
  }
  size*=info.comp;
  try {
    data=new GLubyte [size];
  } catch (std::bad_alloc) {
    Error("Image: not enough memory <-Image::PrepareMem");
  }
}


Image::Image(int width, int height, PixType type)
{
  PrepareMem(width, height, type);
}


Image::Image(std::string filename)
{try{
  File f(filename, "rb");
  BITMAPFILEHEADER bmf;
  BITMAPINFOHEADER bmi;
  if (f.Read(&bmf, sizeof(bmf)) < sizeof(bmf))
    Error(filename+": non BMP file (no file header)");
  if (bmf.bfType==0x4D42) //bm
  {
    if (f.Read(&bmi, sizeof(bmi)) < sizeof(bmi))
      throw Error(filename+": non BMP file (no image header)");
    if (bmf.bfType != ('B'+(short('M')<<8)))
      throw Error(filename+": non BMP file");
    if (bmi.biBitCount!=24)
      throw Error(filename+": can load only 24-bit bmps");
    PrepareMem(bmi.biWidth, bmi.biHeight, RGB);
    int size=info.width*info.height;
    if (f.Read(data, 3*size)<3*size)
    {
      delete data;
      data= 0;
      throw Error(filename+": non BMP file (no data)");
    }
    for (int i=0; i<size; i++) {
      //!opt by asm
      GLubyte t=data[3*i];
      data[3*i]=data[3*i+2];
      data[3*i+2]=t;
    }
  } else if (bmf.bfType==0x4E42) //my gray scale
  {
    if (f.Read(&bmi, sizeof(bmi)) < sizeof(bmi))
      throw Error(filename+": non BN file (no image header)");
    if (bmi.biClrUsed != 256)
      throw Error(filename+": colortable is not 256 length");
    char *tmp=new char[256*4];
    if (f.Read(tmp, 256*4) < sizeof(bmi))
    {
      delete tmp;
      throw Error(filename+": non BN file (no colortable)");
    }
    PrepareMem(bmi.biWidth, bmi.biHeight, ALPHA);
    int size=info.width*info.height;
    if (f.Read(data, size)<size)
    {
      delete data;
      data= 0;
      throw Error(filename+": non BMP file (no data)");
    }
  }
} catch (Error err) {
  err.msg+= " <-Image::Image";
  throw err;
}}


GLuint Image::TurnToGl()
{try{
  GLuint id;
  glGenTextures(1, &id);
  glBindTexture(GL_TEXTURE_2D, id);
  unsigned int i;
  i=info.width;  while (!(i&1)) i>>=1; if (i!=1) Error("Invalid width");
  i=info.height; while (!(i&1)) i>>=1; if (i!=1) Error("Invalid height");
  gluBuild2DMipmaps(
    GL_TEXTURE_2D,
    info.comp,
    info.width,
    info.height,
    info.type,
    GL_UNSIGNED_BYTE,
    data
  );
  glBindTexture(GL_TEXTURE_2D, 0);//?
  delete[] data;
  data=0;
  return id;
} catch (Error err) {
  err.msg+= " <-Image::TurnToGl";
  throw err;
}}

void InterlaceColorAlpha(const Image *from_c, const Image *from_a, Image *to)
{try{
  int w=to->info.width;
  int h=to->info.height;
  if ( //not same width and height. Also define 'w' and 'h'
    ( w != from_c->info.width  || w != from_a->info.width ||
      h != from_c->info.height  || h != from_a->info.height
    )
  )
    throw Error("invalid width or height");
  using Image::RGB;
  using Image::RGBA;
  using Image::ALPHA;
  if (
    RGB == from_c->info.type &&
    ALPHA == from_a->info.type &&
    RGBA == to->info.type
  )
  {
    const GLubyte (*prgb)[3]=
      reinterpret_cast<GLubyte (*)[3]> (from_c->data);
    const GLubyte *pa=
      from_a->data;
    typedef unsigned int uint32;
    uint32 *prgba=
      reinterpret_cast<uint32 *> (to->data);
    for (int i=0; i<w; i++)
    {
      for (int j=0; j<h; j++)
      {
        int rgba=
          int((*prgb)[0]) |
          (int((*prgb)[1]) << 8) |
          (int((*prgb)[2]) << 16) |
          *pa<<24;
        *prgba++ = rgba;
        pa++;
        prgb++;
      }
    }
  }
  else
    throw Error("not implemented");
} catch (Error err) {
  err.msg+= " <-Interlace";
  throw err;
}}

void Convert(const Image *from, Image *to)
{try{
  int w,h;
  if ( //not same width and height. Also define 'w' and 'h'
    (
      ((w=from->info.width) - to->info.width) |
      ((h=from->info.height) - to->info.height)
    )
  )
    throw Error("invalid width or height");
  using Image::RGB;
  using Image::RGBA;
  if (from->info.type ==  to->info.type)
  {
    std::memcpy(to->data, from->data, w*h* to->info.comp);
  }
  /*else if (
    RGB == from->info.type &&
    RGBA == to->info.type
  )
  {

  }*/
  else
    throw Error("not implemented");
} catch (Error err) {
  err.msg+= " <-Convert";
  throw err;
}}


