// (c) Ivan Yanikov, dobrokot 'at' gmail.com

#ifndef IMAGE_LOAD_H_
#define IMAGE_LOAD_H_

#include <string>
#include <windows.h>
#include <GL/gl.h>

// (*to) width and height should be same as (*from)
class Image;
void Convert(const Image *from, Image *to);
void InterlaceColorAlpha(const Image *from_c, const Image *from_a, Image *to);

class Image
{
  friend void Convert(const Image *from, Image *to);
  friend void InterlaceColorAlpha(const Image *from_c, const Image *from_a, Image *to);
  public:
    enum PixType
    {
      RGB=GL_RGB,
      RGBA=GL_RGBA,
      ALPHA=GL_ALPHA,
      LUMA=GL_LUMINANCE,
      LUMA_ALPHA=GL_LUMINANCE_ALPHA
    };
  protected:
    GLubyte *data;
    struct Info
    {
      int width, height;
      PixType type;
      int comp;
    };
    Info info;
    void PrepareMem(int width, int height, PixType type);
  public:
    Image() { data=0; };
    Image(int width, int height, PixType type);
    Image(std::string filename);
    ~Image() { delete data; };
    GLubyte *GetData() { return data; };
    GLubyte *Pix(int j, int i) { return data+info.comp*(i*info.width+j); };
    Info GetInfo() { return info; };
    GLuint TurnToGl();
};

#endif