// (c) 2003, Ivan Yanikov, dobrokot 'at' gmail.com 

class Mesh
{
  struct FileHead
  {
    int obj_type;
    int size_of_obj;
    int num_pts;
    int num_tex_pts;
    int num_trias;
  };
  struct FileVert
  {
    int point;
    int texpt;
    float nx,ny,nz;
  };

  struct FileTria
  {
    FileVert vx[3];
    int TexId;
  };

  struct Filep3d
  {
    float x,y,z;
  };

  struct Filep2d
  {
    float x,y;
  };
  protected:
    float time;
    std::vector<float3> points;
    std::vector<float3> texps;
    std::vector<float3> normals;
    struct Triangle
    {
        int p[3];
        int n[3];
        int e[3];
        int t[3];
    };
    std::vector<Triangle> trs;
  public:
    GLuint tex;
    Matrix pos;
    void LoadFromFile(std::string file);
    virtual void GetTime(double time) {this->time=time;};
    virtual void Draw();
};


void Mesh::Draw()
{
  using namespace std;
  glBindTexture(GL_TEXTURE_2D, tex);
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  pos.glMult();
  glBegin(GL_TRIANGLES);
  for (vector<Triangle>::iterator tr_i=trs.begin();
       tr_i!=trs.end();
       ++tr_i)
  {
    glNormal3fv(normals[tr_i->n[0]].xyz);
    glTexCoord2fv(texps[tr_i->t[0]].xyz);
    glVertex3fv(points[tr_i->p[0]].xyz);

    glNormal3fv(normals[tr_i->n[1]].xyz);
    glTexCoord2fv(texps[tr_i->t[1]].xyz);
    glVertex3fv(points[tr_i->p[1]].xyz);

    glNormal3fv(normals[tr_i->n[2]].xyz);
    glTexCoord2fv(texps[tr_i->t[2]].xyz);
    glVertex3fv(points[tr_i->p[2]].xyz);
  }
  glEnd();
  glPopMatrix();
}


void Mesh::LoadFromFile(std::string file)
{
  File mdl(file+".mdl", "rb");
  FileHead fh;
  mdl.ReadSmart(fh);
  points.resize(fh.num_pts);
  texps.resize(fh.num_tex_pts);
  trs.resize(fh.num_trias);
  normals.resize(3*fh.num_trias);
  Filep3d *f_points=new Filep3d[points.size()];
  Filep2d *f_texps=new Filep2d[texps.size()];
  FileTria *f_trs=new FileTria[trs.size()];
  mdl.ReadSmart(*f_points, points.size());
  mdl.ReadSmart(*f_texps, texps.size());
  mdl.ReadSmart(*f_trs, trs.size());
  for (unsigned i=0; i<points.size(); i++) {
    points[i].x=f_points[i].x;
    points[i].y=f_points[i].z;
    points[i].z=f_points[i].y; //3ds - other axis
  }
  for (unsigned i=0; i<texps.size(); i++) {
    texps[i].x=f_texps[i].x/256;
    texps[i].y=f_texps[i].y/256;
    texps[i].z=0;
  }
  for (unsigned i=0; i<trs.size(); i++) {
    for (int j=0; j<3; j++)
      trs[i].p[j]=f_trs[i].vx[j].point;
    for (int j=0; j<3; j++)
      trs[i].t[j]=f_trs[i].vx[j].texpt;
    for (int j=0; j<3; j++) {
      trs[i].n[j]=3*i+j;
      normals[3*i+j].x=f_trs[i].vx[j].nx;
      normals[3*i+j].y=f_trs[i].vx[j].ny;
      normals[3*i+j].z=f_trs[i].vx[j].nz;
    };
  };
  delete f_trs;
  delete f_texps;
  delete f_points;
}




