// (c) 2003, Ivan Yanikov, dobrokot 'at' gmail.com 
#include <gl_clock.h>

//--------------------  class Animation ------------------------------

void Animation::Init(std::string *istr, Application *app)
{try{
  this->appl=app;
  state.phase=0;
  state.t=0;
  cur_time=0;
  forward=false;
  at_begin=true;
  at_end=false;
  if (istr)                                
  {
    using namespace std;
    ifstream file((*istr+".ani").c_str());
    if (!file)
      throw Error(string("Can't open file ") + *istr);
    file.exceptions(std::ios_base::failbit);
    try
    {
      file >>n;
      float t=0;
      for (int i=0; i<n; i++)
      {
        float dt;
        file >>dt;
        keys.push_back(t+=dt);
      }
      total_time=t;
    }
    catch (std::ios_base::failure)
    {
      throw Error(std::string("input error (file ")+*istr+")");
    }
  }
  else
  { //!? at_begin=true?
    total_time=0;
    n=0;
  }
} catch (Error err) {
  err.msg+= " <-Animation::Init";
  throw err;
}}


void Animation::NextFrame()
{
  float dt = appl->time - appl->lf_time;
  if (forward)
  {
    if (!at_end)
    {
      at_begin=false;
      cur_time += dt;
      int phase=state.phase;
      while (phase<n && cur_time>keys[phase])
        phase++;
      if (phase==n)
      {
        cur_time=total_time;
        state.phase=n-1;
        state.t=1.0;
        at_end=true;
      } else {
        state.phase=phase;
        float prev_key;
        if (phase==0)
          prev_key=0;
        else
          prev_key=keys[phase-1];
        state.t=(cur_time-prev_key)/(keys[phase]-prev_key);
      }
    }
  }
  else // !forward
  {
    if (!at_begin)
    {
      at_end=false;
      cur_time -= dt;
      int phase=state.phase;
      while (phase>0 && cur_time<=keys[phase-1]) //!? []> or >=
        phase--;
      if (cur_time<0) //  phase==0
        phase--; //phase=-1;
      if (phase==-1)
      {
        cur_time=0;
        state.phase=0;
        state.t=0;
        at_begin=true;
      } else {
        state.phase=phase;
        float prev_key;
        if (phase==0)
          prev_key=0;
        else
          prev_key=keys[phase-1];
        state.t=(cur_time-prev_key)/(keys[phase]-prev_key);
      }
    }
  }
}


float Animation::t(int phase) const
{
  if (phase == state.phase)
    return state.t;
  else if (phase < state.phase)
    return 1.0f;
  else
    return 0.0f;
}

//-----------------  class AniMeshObj -------------

void AniMeshObj::Init(std::string istr, Application *app)
{try{
  //if (istr)
  {
    using namespace std;
    ifstream file((istr+".aniobj").c_str());
    if (!file)
      throw Error(string("Can't open file ") + istr);
    file.exceptions(std::ios_base::failbit);
    try
    {
      string meshfile;
      string anifile;
      file >>meshfile;
      file >>anifile;
      MeshObj::Init(meshfile, app);
      Animation::Init(&anifile, app);
    }
    catch (std::ios_base::failure)
    {
      throw Error(std::string("input error (file ")+istr+")");
    }
  }  
}catch (Error err) {
  err.msg+= " <-AniMeshObj ::Init";
  throw err;
}}

//-----------------  class Obj  -----------------

void Obj::Init(std::string istr, Application *app)
{
  this->app = app;
  /*
  for (
    std::vector <Obj *>::iterator obj_i=sub_objs.begin();
    obj_i<sub_objs.end();
    obj_i++
  ) {
    (**obj_i).SetApplication(app);
  }*/
}

void Obj::Draw(bool select)
{
  for (
    std::vector <Obj *>::iterator obj_i=sub_objs.begin();
    obj_i<sub_objs.end();
    obj_i++
  ) {
    (**obj_i).Draw(select);
  }
}


//----------------  class MeshObj  --------------------------------------
void MeshObj::Init(std::string istr, Application *app)
{try{
  Obj::Init(istr, app);
  std::string mdlfile;
  std::string texfile;
  float diffuse_col[4];
  bool is_tex;
  try
  {
    istr+=".mshobj";
    std::ifstream objfile(istr.c_str());
    if (!objfile)
      throw Error(std::string("can't open file ") + istr);
    objfile.exceptions(std::ios_base::failbit);
    objfile >>mdlfile >>texfile;
    is_tex = (texfile!="no");
    if (!is_tex)
    {
      objfile >>diffuse_col[0];
      objfile >>diffuse_col[1];
      objfile >>diffuse_col[2];
      objfile >>diffuse_col[3];
    }
  }
  catch (std::ios_base::failure)
  {
    throw Error(std::string("input error (file ")+istr+")");
  }
  File mdl(mdlfile, "rb");
  FileHead fh;
  mdl.ReadSmart(fh);
  Filep3d *f_points=new Filep3d[fh.num_pts];
  Filep2d *f_texps=new Filep2d[fh.num_tex_pts];
  FileTria *f_trs=new FileTria[fh.num_trias];
  mdl.ReadSmart(*f_points, fh.num_pts);
  mdl.ReadSmart(*f_texps, fh.num_tex_pts);
  mdl.ReadSmart(*f_trs, fh.num_trias);

  list = glGenLists(1);
  glNewList(list, GL_COMPILE);
  if (!is_tex)
  {
    glMaterialfv(GL_FRONT, GL_DIFFUSE, diffuse_col);
  }

  glBegin(GL_TRIANGLES);
  for (int i=0; i<fh.num_trias; i++)
  {
    for (int j=0; j<3; j++)
    {
      if (is_tex)
      {
        glTexCoord2f(
          f_texps[f_trs[i].vx[j].texpt].x/256,
          f_texps[f_trs[i].vx[j].texpt].y/256
        );
      }
      glNormal3f(
        f_trs[i].vx[j].nx,
        f_trs[i].vx[j].ny,
        f_trs[i].vx[j].nz
      );
      glVertex3f(
        f_points[f_trs[i].vx[j].point].x,
        f_points[f_trs[i].vx[j].point].y,
        f_points[f_trs[i].vx[j].point].z
      );
    }
  };
  glEnd();
  if (!is_tex)
  {
    float white_diffuse_col[4]={1,1,1,1};
    glMaterialfv(GL_FRONT, GL_DIFFUSE, white_diffuse_col);
  }
  /*//show normals:
  glDisable(GL_LIGHTING);
  glBegin(GL_LINES);
  glColor3f(0,1,0);
  for (int i=0; i<fh.num_trias; i++)
  {
    for (int j=0; j<3; j++)
    {
      glVertex3f(
        f_points[f_trs[i].vx[j].point].x,
        f_points[f_trs[i].vx[j].point].y,
        f_points[f_trs[i].vx[j].point].z
      );
      glVertex3f(
        f_points[f_trs[i].vx[j].point].x+1*f_trs[i].vx[j].nx,
        f_points[f_trs[i].vx[j].point].y+1*f_trs[i].vx[j].ny,
        f_points[f_trs[i].vx[j].point].z+1*f_trs[i].vx[j].nz
      );
    }
  }
  glEnd();
  glEnable(GL_LIGHTING);/**/
  glEndList();
  tex=0;
  if (is_tex)
  {
    Image image(texfile);
    tex=image.TurnToGl();
  }
  delete[] f_trs;
  delete[] f_texps;
  delete[] f_points;
} catch (Error err) {
  err.msg+=" <-MeshObj::Init";
  throw err;
}}


void MeshObj::Draw(bool select)
{
  if (!tex)
    glDisable(GL_TEXTURE_2D);
  else
    glBindTexture(GL_TEXTURE_2D, tex);
  glCallList(list);

  if (!tex)
    glEnable(GL_TEXTURE_2D);
}


MeshObj::~MeshObj()
{
  glDeleteLists(list, 1);
  glDeleteTextures(1, &tex);
}



//----------------   class Scene  --------------------------------------


void Scene::Init(std::string istr, Application *app)
{
  Obj::Init(istr, app);
  cam_rot=1;
  cam_dist=33;
  cam_vx=0;
  cam_vy=0;
  //obj.Init("lala");
  cherep.Init("", app);
  sub_objs.push_back(&cherep);
}


void Scene::Draw(bool select)
{
  {
    cam_rot=
      Matrix::Rotatey(cam_vx)*
      Matrix::Rotatex(cam_vy)*
      cam_rot;
    cam_vx*=0.90;
    cam_vy*=0.90;
  }
  //static in=0;
  //if (!in) {
  //}
  if (app->mouse_right)
  {
    app->mouse_right=false;
    app->CheckHits();
    return;
  }
  glClearColor(0, 0, 0.2, 0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  if (app->height != 0) {
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);

    gluPerspective(60, float(app->width) / float(app->height), 1, 1000);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    app->cur_matrix=1;

    float pos[4]={1, 1, 4};
    glLightfv(GL_LIGHT0, GL_POSITION, pos);

    app->Mult(
      Matrix::Translate(0, 0, -cam_dist)*
      cam_rot
    );

    /*glBegin(GL_TRIANGLES);
      glNormal3f(0, 0, 1);
      glVertex3f(10, 10, 0);
      glVertex3f(0, 10, 0);
      glVertex3f(10, 0, 0);
    glEnd();/**/
    Obj::Draw(select);
    glDisable(GL_LIGHT0);
    glDisable(GL_LIGHTING);
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_TEXTURE_2D);
  }
}


void Scene::UpdateMouse(int dx, int dy)
{
  if (app->mouse_left)
  {
    cam_vx+=0.8*float(dx)/app->width;
    cam_vy+=0.8*float(dy)/app->height;
  }
}


//-----------------  struct Particle  -----------------------------

inline void ParticleRing::Particle::Draw(Matrix &m)
{
  //MODELVIEW in gl should be identity
  //m should be real MODELVIEW.
  //OpenGL should be in glBegin(GL_QUADS) mode

  float3 centr = m*pos;
  float size_2=size*0.5;

  centr.x+=size_2;
  centr.y+=size_2;
  glTexCoord2f(1,1);
  glVertex3fv(&centr.x);
  centr.x-=size;
  glTexCoord2f(0,1);
  glVertex3fv(&centr.x);
  centr.y-=size;
  glTexCoord2f(0,0);
  glVertex3fv(&centr.x);
  centr.x+=size;
  glTexCoord2f(1,0);
  glVertex3fv(&centr.x);
};


/*class ParticleRing
  :public Obj
{
    GLuint texid;
    Particle prs;
  public:
    virtual void Init(std::string istr, Application *app);
    inline void Draw(bool select);
    virtual ~ParticleRing();
};*/

void ParticleRing::Init(std::string istr, Application *app)
{
  //texid = Image("ParticleRing").TurnToGl();
  Obj::Init(istr, app);
  struct Rgba
  {
    GLubyte r,g,b,a;
  };
  const tex_size=16;
  typedef Rgba strip[tex_size];
  strip *tex = new strip[tex_size];
  for (int i=0; i<tex_size; i++)
  {
    for (int j=0; j<tex_size; j++)
    {
      tex[i][j].r=0;
      tex[i][j].g=160;
      tex[i][j].b=50;
      int jj= j-tex_size/2;
      int ii= i-tex_size/2;
      int kk= tex_size*tex_size/4 - jj*jj - ii*ii;
      if (kk>0)
        tex[i][j].a= GLubyte(60*float(kk) / (tex_size*tex_size/4));
      else
        tex[i][j].a=0;
    }
  }
  glGenTextures(1, &texid);
  glBindTexture(GL_TEXTURE_2D, texid);
  gluBuild2DMipmaps(
    GL_TEXTURE_2D,
    4,
    tex_size,
    tex_size,
    GL_RGBA,
    GL_UNSIGNED_BYTE,
    tex
  );
  delete[] tex;

  n=200;
  num_speeds=10;

  prs= new pParticle[num_speeds];
  for (int speed=0; speed < num_speeds; speed++)
  {
    prs[speed]= new Particle[n];
    for (int i=0; i<n; i++)
    {
      using namespace std;
      float phi= 3.14159*2*float(std::rand())/RAND_MAX;
      float teta = 3.14159*2*float(std::rand())/RAND_MAX;
      float r=3+0.1*float(std::rand())/RAND_MAX;
      float rr=12;
      prs[speed][i].size= 2*float(std::rand())/RAND_MAX;

      prs[speed][i].pos.x= (rr+r*cos(phi))*cos(teta);
      prs[speed][i].pos.y= sin(phi)*r-4;
      prs[speed][i].pos.z= (rr+r*cos(phi))*sin(teta);
    }
  }
}


void ParticleRing::Draw(bool select)
{
  if (!select)
  {
    Matrix matrix_save;
    app->Save(&matrix_save);
    app->Mult(
      Matrix::Translate(0, 0, -3.5)
      //Matrix::Rotatex(3.14159/2)
    );

    glDisable(GL_LIGHTING);
    glEnable(GL_BLEND);


      glDepthMask(GL_FALSE);
      glBlendFunc(GL_SRC_ALPHA, GL_ONE);
      glBindTexture(GL_TEXTURE_2D, texid);

      glColor4f(1, 1, 1, transp);
      const float delta_ug=0.1;
      app->Mult(
        Matrix::Rotatey(-delta_ug*app->time*num_speeds/2)
      );
      for (int speed=0; speed <num_speeds; speed++)
      {
          app->Mult(
            Matrix::Rotatey(delta_ug*app->time)
          );
          glLoadIdentity();
          glBegin(GL_QUADS);
            //make cur_matrix LOCAL variable for optimization
            Matrix m = app->cur_matrix;
              int i=0;
              for (pParticle p_p=prs[speed]; i<n; i++, p_p++)
                p_p->Draw(m);
          glEnd();
      }
      glDepthMask(GL_TRUE);
    glDisable(GL_BLEND);
    glEnable(GL_LIGHTING);
    app->Restore(&matrix_save);
  }
}


ParticleRing::~ParticleRing()
{
  glDeleteTextures(1, &texid); //! MeshObj no delete!
  for (int speed=0; speed <num_speeds; speed++)
  {
    delete[] prs[speed];
  }
  delete[] prs;
}


//----- class WatchFace, Zuben, Buben, Clock --------------------------------



WatchFace::~WatchFace()
{
  glDeleteTextures(1, &texid);
}


void WatchFace::Init(std::string istr, Application *app)
{
  Obj::Init(istr, app);
  min_arrow.Init("min_arrow", app);
  cl_arrow.Init("cl_arrow", app);
  sub_objs.push_back(&cl_arrow);
  sub_objs.push_back(&min_arrow);

  Image alpha("face_mask.bmp");
  Image rgb("face.bmp");
  Image rgba(512, 512, Image::RGBA);
  InterlaceColorAlpha(&rgb, &alpha, &rgba);
  texid = rgba.TurnToGl();
  transp = 1;
}


void WatchFace::Draw(bool select)
{
  glBindTexture(GL_TEXTURE_2D, texid);
  float3 centr(0, -2, 5);
  float size=20;
  int n=20;
  //float size_2=size*0.5;

  glEnable(GL_BLEND);
  glEnable(GL_ALPHA_TEST);
    glDisable(GL_LIGHTING);
      glAlphaFunc(GL_GREATER, 0.1);
      //glDepthMask(GL_FALSE);
      glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
      glBindTexture(GL_TEXTURE_2D, texid);
      glColor4f(1, 1, 1, transp);
      for (int i=0; i<n; i++)
      {
        glBegin(GL_QUAD_STRIP);
        for (int j=0; j<=n; j++)
        {
          for (int k=0; k<2; k++)
          {
            float u=(i+k)/float(n);
            float v=(j)/float(n);
            glTexCoord2f(u,v);
            glVertex3f(
              centr.x-size/2.0+u*size+(1.2-transp)*0.55*std::sin(((i+k)^j)+5*app->time),
              centr.y-size/2.0+v*size+(1.2-transp)*0.55*std::cos(((i+k)^j)+5*app->time),
              centr.z
            );
          }
        }
        glEnd();
      }
      /*glBegin(GL_QUADS);
        centr.x+=size_2;
        centr.y+=size_2;
        glTexCoord2f(1,1);
        glVertex3fv(&centr.x);
        centr.x-=size;
        glTexCoord2f(0,1);
        glVertex3fv(&centr.x);
        centr.y-=size;
        glTexCoord2f(0,0);
        glVertex3fv(&centr.x);
        centr.x+=size;
        glTexCoord2f(1,0);
        glVertex3fv(&centr.x);
      glEnd();*/
    glEnable(GL_LIGHTING);
      float diff[4] = {1.0, 0.8, 0.56, transp};
      glMaterialfv(GL_FRONT, GL_DIFFUSE, diff);
      Matrix matrix_save;
      app->Save(&matrix_save);
      app->Mult(
        Matrix::Translate(0, -2, 8.5) *
        Matrix::Rotatex(3.14159/2)*
        Matrix::Rotatey(
          -(app->loc_time.tm_hour + app->loc_time.tm_hour/60.0)/12.0*3.14159*2
        )

      );
      cl_arrow.Draw(select);
      app->Restore(&matrix_save);

      matrix_save;
      app->Save(&matrix_save);
      app->Mult(
        Matrix::Translate(0, -2, 8.5) *
        Matrix::Rotatex(3.14159/2) *
        Matrix::Rotatey(-app->loc_time.tm_min/60.0*3.14159*2)
      );
      min_arrow.Draw(select);
      app->Restore(&matrix_save);
  glDisable(GL_ALPHA_TEST);
  //glDepthMask(GL_TRUE);
  glDisable(GL_BLEND);
}


void Eye::OnSelect(void *)
{
  forward= !forward;
}


void Eye::Init(std::string istr, Application *app)
{
  AniMeshObj::Init(istr, app);
  name= app->CallMe(this, 0);
}


void RightEye::Draw(bool select)
{
  NextFrame();
  Matrix matrix_save;
  app->Save(&matrix_save);
    app->Mult(
      Matrix::Translate(-2.208, -4.184-4*std::sin(t(0)*3.14159), 1.646)*
      Matrix::Rotatez(0.4*std::sin(app->time*2*3.14159))*
      Matrix::Rotatex(0.2*std::sin(app->time))
    );
    glPushName(name);
      MeshObj::Draw(select);
    glPopName();
    Obj::Draw(select);
  app->Restore(&matrix_save);
}

void LeftEye::Draw(bool select)
{
  NextFrame();
  Matrix matrix_save;
  app->Save(&matrix_save);
    {
      float spec[4]={1.5,1.5,1.5,1};
      glMaterialfv(GL_FRONT, GL_SPECULAR, spec);
      glMaterialf(GL_FRONT, GL_SHININESS, 10);
    }
    app->Mult(
      Matrix::Translate(2.208, -4.184-4*std::sin(t(0)*3.14159), 1.646)*
      Matrix::Rotatez(0.4*std::sin(app->time*2*3.14159))*
      Matrix::Rotatex(-0.21*std::sin(app->time))
    );
    glPushName(name);
      MeshObj::Draw(select);
    glPopName();
    {
      float spec[4]={0,0,0,1};
      glMaterialfv(GL_FRONT, GL_SPECULAR, spec);
    }
    Obj::Draw(select);
  app->Restore(&matrix_save);
}


/*    virtual void Init(std::string istr, Application *app);
    virtual void SelectDraw();
    virtual void Draw(bool select);*/

void Maniac::Draw(bool select)
{
  //glPushMatrix();
  Matrix matrix_save;
  app->Save(&matrix_save);
    //glTranslatef(0, 0, -5);
    app->Mult(
      Matrix::Translate(0, 0, -5)*
      Matrix::Rotatey(std::sin(app->time*3.14159)/3.0)
    );
    //glRotatef(std::sin(app->time*3.14159)*20, 0, 1, 0);
    MeshObj::Draw(select);
    Obj::Draw(select);
  //glPopMatrix();
  app->Restore(&matrix_save);
}


void Zuben::Init(std::string istr, Application *app)
{
  MeshObj::Init(istr, app);
  maniac.Init("maniac", app);
  sub_objs.push_back(&maniac);
  this->name=app->CallMe(this, 0);

}


void Zuben::OnSelect(void *)
{
  MessageBox(
    0,
    "Access denied.    .    -      .",
    "Operation is not permitted.",
    MB_OK
  );
}


void Zuben::Draw(bool select)
{
  //glPushMatrix();
  Matrix matrix_save;
  app->Save(&matrix_save);
  glPushName(name);
    //glRotatef((std::sin(app->time*3)+1)*20, 1, 0, 0);
    //glRotatef(std::sin(app->time*10)*5, 0, 0, 1);
    MeshObj::Draw(select);
    Obj::Draw(select);
  glPopName();
  glPopMatrix();
  app->Restore(&matrix_save);
}



void Buben::Init(std::string istr, Application *app)
{
  AniMeshObj::Init(istr, app);
  reye.Init("eye", app);
  leye.Init("eye", app);
  sub_objs.push_back(&reye);
  sub_objs.push_back(&leye);
  white=false;
  this->name=app->CallMe(this, 0);
}


void Buben::Draw(bool select)
{
  NextFrame();
  Matrix matrix_save;
  app->Save(&matrix_save);
    app->Mult(
      Matrix::Rotatex(-1.4*t(0))*
      Matrix::Rotatez(0.015*std::sin(app->time*2*3.14159))
    );
    app->scene.cherep.particle_ring.transp=t(0);
    app->scene.cherep.watch_face.transp=t(1);

    //glDisable(GL_TEXTURE_2D);
    glPushName(name);
      MeshObj::Draw(select);
    glPopName();
    //glEnable(GL_TEXTURE_2D);

    Obj::Draw(select);
  app->Restore(&matrix_save);
}


void Buben::OnSelect(void *)
{
  white = !white;
  forward= !forward;
}

void Clock::Init(std::string istr, Application *app)
{
  Obj::Init(istr, app);
  buben.Init("buben", app);
  zuben.Init("zuben", app);
  watch_face.Init("", app);
  particle_ring.Init("", app);
  sub_objs.push_back(&buben);
  sub_objs.push_back(&zuben);
  sub_objs.push_back(&watch_face);

  //transparent:
  sub_objs.push_back(&particle_ring);
}


void Clock::Draw(bool select)
{
  Matrix matrix_save;
  app->Save(&matrix_save);
    //glRotatef(-90, 1, 0, 0);
    app->Mult(
      Matrix::Rotatex(-1.57)*
      Matrix::Translate(0, 0, 2)
    );
    Obj::Draw(select);
  app->Restore(&matrix_save);
}


//----------------   class Application  --------------------------------------

Application::Application()
{
  name_buffer=0;
}


void Application::Init(int argc, char* argv[])
{try{
  void OnDisplay();
  void OnKey(unsigned char, int, int);
  void OnSize(int width, int height);
  void OnMouseMove(int x, int y);
  void OnMouseButton(int button, int state, int x, int y);
  void OnIdle();

  name_buffer_size=512;
  name_buffer=new GLuint[name_buffer_size];

  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
  glutInitWindowSize(640, 480);
  glutCreateWindow("Clocks of the Evil");

  //glutFullScreen();
  //fullscreen = true;
  fullscreen = false;

  glutDisplayFunc(OnDisplay);
  glutKeyboardFunc(OnKey);
  glutReshapeFunc(OnSize);
  glutMotionFunc(OnMouseMove);
  glutPassiveMotionFunc(OnMouseMove);
  glutMouseFunc(OnMouseButton);
  glutIdleFunc(OnIdle);

  scene.Init("", this);

  mouse_moved=false;

  QueryPerformanceCounter(
    (LARGE_INTEGER *)&start_performance_counter
  );
  QueryPerformanceFrequency(
    (LARGE_INTEGER *)&performance_frequency
  );
  lf_time=0;
} catch (Error err) {
  err.msg+=" <-Application::Init";
  throw err;
}}


Application::~Application()
{
  delete[] name_buffer;
  glutDestroyWindow(glutGetWindow());
}


GLuint Application::CallMe(Obj* obj, void *param)
{
  NameRec name_rec={obj, param};
  name_storage.push_back(name_rec);
  return name_storage.size()-1;
}


void Application::CheckHits()
{
  glSelectBuffer(name_buffer_size, name_buffer);
  glRenderMode(GL_SELECT);
  glInitNames();

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  int viewport[4]= {0, 0, width, height};
  gluPickMatrix(
    old_mouse_x, height - old_mouse_y,
    2, 2,
    viewport
  );

  scene.Draw(true);
  //scene.Draw();
  //return;

  GLuint nhits= glRenderMode(GL_RENDER);
  GLuint i, j;
  GLuint *ptr= name_buffer;
  GLuint zmin= ~0;
  int zmin_name=-1;
  for(i= 0; i<nhits; i++ )
  {
    GLuint name_stack_depth= *ptr++;
    GLuint z=*ptr++; //zmin
    ptr++; //zmin
    for(j= 0; j<name_stack_depth; j++)
    {
      GLuint currname= *ptr++;
      if (zmin >= z)
      {
        zmin=z;
        zmin_name=currname;
      }
      //name_storage[currname].obj->
        //OnSelect(name_storage[currname].param);
    }
  }
  if (zmin_name!=-1)
    name_storage[zmin_name].obj->
      OnSelect(name_storage[zmin_name].param);
}


inline void Application::Mult(const Matrix &mult)
{
  cur_matrix=cur_matrix*mult;
  mult.glMult();
}


inline void Application::Save(Matrix *save_space)
{
  depth++;
  glPushMatrix();
  *save_space = cur_matrix;
}


void Application::Restore(const Matrix *save_space)
{
  cur_matrix = *save_space;
  glPopMatrix();
  depth--;
}


void Application::OnKey(unsigned char key, int x, int y)
{
  if (key == 033)
    std::exit(0);
    //!throw Exit();
  if (key=='\r')
  {
    fullscreen=!fullscreen;
    if (fullscreen)
      glutFullScreen();
    else
    {
      glutPositionWindow(0, GetSystemMetrics(SM_CYCAPTION));
      glutReshapeWindow(GetSystemMetrics(SM_CXFULLSCREEN), GetSystemMetrics(SM_CYFULLSCREEN));
    }
  }
}


void Application::OnDisplay()
{try{

  __int64 current_performance_counter;
  QueryPerformanceCounter(
    (LARGE_INTEGER *)&current_performance_counter
  );
  current_performance_counter-=start_performance_counter;
  time= (long double)current_performance_counter/performance_frequency;

  std::time_t gmt_sec= std::time(0);
  loc_time= *std::localtime(&gmt_sec);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  depth = 0;
  scene.Draw(false);
  if (depth != 0)
    throw Error("invalid matirx stack depth", depth);
  lf_time= time;
  glutSwapBuffers();
} catch (Error err) {
  err.msg+=" <-Application::OnDisplay";
  throw err;
}}



void Application::OnSize(int width, int height)
{
  this->width=width;
  this->height=height;
  glViewport(0, 0, width, height);
}


void Application::OnMouseMove(int x, int y)
{
  if (mouse_moved)
      scene.UpdateMouse(x-old_mouse_x, y-old_mouse_y);
  else
    mouse_moved=true;
  old_mouse_x=x;
  old_mouse_y=y;
}

void Application::OnMouseButton(int button, int state, int x, int y)
{
  switch (button)
  {
    case GLUT_LEFT_BUTTON:
      mouse_left=(state==GLUT_DOWN);
      break;
    case GLUT_RIGHT_BUTTON:
      mouse_right=(state==GLUT_DOWN);
      break;
    case GLUT_MIDDLE_BUTTON:
      mouse_middle=(state==GLUT_DOWN);
  }
}


void Application::OnIdle()
{
  glutPostRedisplay();
}


void Application::Run()
{
  try {
    glutMainLoop();
  }
  catch (Exit)
  {
  }
}

Application application;

void OnKey(unsigned char key, int x, int y)
{
  application.OnKey(key, x, y);
}

void OnSize(int width, int height)
{
  application.OnSize(width, height);
}

void OnDisplay()
{
  application.OnDisplay();
}

void OnMouseMove(int x, int y)
{
  application.OnMouseMove(x, y);
}

void OnMouseButton(int button, int state, int x, int y)
{
  application.OnMouseButton(button, state, x, y);
}

void OnIdle()
{
  application.OnIdle();
}

#pragma argsused
int main(int argc, char* argv[])
{try{
  application.Init(argc, argv);
  application.Run();
  return 0;
} catch (Error err) {
  err.msg+=" <-main";
  err.Print();
  return 1;
}}
