// (c) Ivan Yanikov, dobrokot 'at' gmail.com
#include <vector>
#include <cstdlib>
#include <string>
#include <GL/glut.h>
#include <GL/glu.h>
#include <fstream>
#include <iostream>
#include <windows.h>
#include <ctime>
#include "file.h"
#include "error.h"
#include "ImageLoad.h"
#include "matrix.h"
#pragma hdrstop

//--------------------

class Application;
class Obj;
class Scene;

class Obj
{
  protected:
    GLuint name;
    std::vector <Obj *> sub_objs;
    Application *app;
    //int glRotatef, glTranslatef, glScalef, glPopMatric, glPushMatrix;
  public:
    virtual void Init(std::string istr, Application *app);
    //virtual ~Obj();
    virtual void Draw(bool select);
    virtual void SetState() {};
    virtual void OnSelect(void *param) {};
};

class MeshObj:
  public Obj
{
    GLuint list;
    GLuint tex;
    //file sructures:
    struct FileHead
    {
      int obj_type;
      int size_of_obj;
      int num_pts;
      int num_tex_pts;
      int num_trias;
    };
    struct FileVert
    {
      int point;
      int texpt;
      float nx,ny,nz;
    };
    struct FileTria
    {
      FileVert vx[3];
      int TexId;
    };
    struct Filep3d
    {
      float x,y,z;
    };
    struct Filep2d
    {
      float x,y;
    };
  public:
    virtual void Init(std::string istr, Application *app);
    virtual void Draw(bool select);
    ~MeshObj();
};

//----------------  class Animation ---------------------------

class Animation
{
  public:
    void Init(std::string *istr, Application *app);
    int n; //r //number of phases
    bool AtEnd() {return at_end;};
    bool AtBegin() {return at_begin;};
    bool forward; //rw
    int AddKey(float delta_time);
    void PlayReverse();
    struct State
    {
      int phase;
      float t;
    };
    State GetState() {return state;};
    void NextFrame();
    float t(int phase) const;
  private:
    bool at_end;
    bool at_begin;
    Application *appl;
    State state;
    float total_time;
    float cur_time;
    std::vector <float> keys;
};

//----------------  AnimatedMeshObj  -------------------------------------

class AniMeshObj:
  public MeshObj, protected Animation
{
  public:
    virtual void Init(std::string istr, Application *app);
};

//----------------  class Clock  declaration and his assistants -----------

class Eye
  :public AniMeshObj
{
  public:
    virtual void Init(std::string istr, Application *app);
    virtual void OnSelect(void *);
};

class RightEye
  :public Eye
{
  public:
    virtual void Draw(bool select);
};


class LeftEye
  :public Eye
{
  public:
    virtual void Draw(bool select);
};


class Maniac
  :public MeshObj
{
  public:
    void Draw(bool select);
};

class Buben
  :public AniMeshObj
{
    bool white;
    RightEye reye;
    LeftEye leye;
  public:
    void Init(std::string istr, Application *app);
    void SetApplication(Application *app);
    void SelectDraw();
    void Draw(bool select);
    virtual void OnSelect(void *);
};

class Zuben
  :public MeshObj
{
    Maniac maniac;
  public:
    void Init(std::string istr, Application *app);
    void Draw(bool select);
    void OnSelect(void *);
};


class ParticleRing
  :public Obj
{
    struct Particle
    {
      float size;
      float3 pos;
      float cos_speed;
      //MODELVIEW in gl should be identity
      //m should be real MODELVIEW.
      //OpenGL should be in glBegin(GL_QUADS) mode
      inline void Draw(Matrix &m);
    };
    GLuint texid;
    int num_speeds;
    int n;
    typedef Particle *pParticle;
    pParticle *prs; //array [0..num_speeds-1] of array [0..n-1] of Particle
  private: friend class Buben;  
    float transp;
  public:
    virtual void Init(std::string istr, Application *app);
    void Draw(bool select);
    virtual ~ParticleRing();
};


class WatchFace
  :public Obj
{
    GLuint texid;
    MeshObj min_arrow;
    MeshObj cl_arrow;
  public:
    float transp;
    void Init(std::string istr, Application *app);
    void Draw(bool select);
    ~WatchFace();
};


class Clock
  :public Obj
{
    Buben buben;
    Zuben zuben;
  public:
    WatchFace watch_face;
    ParticleRing particle_ring;
    void Init(std::string istr, Application *app);
    void Draw(bool select);
};


//------------------------- class Scene declaration --------------------------


class Scene:
  public Obj
{
    MeshObj obj;
    Matrix cam_rot;
    float cam_dist;
    float cam_vx, cam_vy;
  public:
    Clock cherep;
    void Init(std::string istr, Application *app);
    void Draw(bool select);
    void UpdateMouse(int dx, int dy);
};

class Application
{
    bool fullscreen;
    class Exit {}; //throw if ESC pressed
    bool mouse_moved;
    int old_mouse_x, old_mouse_y;
    struct NameRec
    {
      Obj *obj;
      void *param;
    };
    std::vector <NameRec> name_storage;
    GLuint *name_buffer;
    GLsizei name_buffer_size;
    __int64 start_performance_counter;
    __int64 performance_frequency;
  public:
    Scene scene;
    void Init(int argc, char* argv[]);
    Application();
    ~Application();

    void OnKey(unsigned char key, int x, int y);
    void OnDisplay();
    void OnSize(int width, int height);
    void Run();
    void OnMouseMove(int x, int y);
    void OnMouseButton(int button, int state, int x, int y);
    void OnIdle();

    int width; //r
    int height; //r

    long double time; //r
    long double lf_time; //r
    struct std::tm loc_time;

    bool mouse_left; //rw
    bool mouse_middle; //rw
    bool mouse_right; //rw

    GLuint CallMe(Obj* obj, void *param);
    void CheckHits();

    Matrix cur_matrix;  //rw
    int depth; //!d
    inline void Save(Matrix *save_space);
    inline void Mult(const Matrix &mult);
    //void SaveMult(Matrix *save_space, const Matrix &mult);
    inline void Restore(const Matrix *save_space);
    //void DelName(GLuint gl_name);
};



 