// (c) 2003, Ivan Yanikov, dobrokot 'at' gmail.com 
#include <matrix.h>
#include <GL/gl.h>
#pragma hdrstop

float Length(const float3 &p)
{
  return std::sqrt(p.x*p.x+p.y*p.y+p.z*p.z);
}


inline float SqrLength (const float3 &p)
{
  return (p.x*p.x+p.y*p.y+p.z*p.z);
}


inline float3 &operator-=(float3 &p2, const float3 &p1)
{
  p2.x-=p1.x;
  p2.y-=p1.y;
  p2.z-=p1.z;
  return p2;
}


inline float3 operator-(const float3 &p2, const float3 &p1)
{
  return float3(p2.x-p1.x, p2.y-p1.y, p2.z-p1.z);
}


inline float3 operator-(const float3 &p)
{
  return float3(-p.x, -p.y, -p.z);
}


inline float3 &operator+=(float3 &p1,const float3 &p2)
{
  p1.x+=p2.x;
  p1.y+=p2.y;
  p1.z+=p2.z;
  return p1;
}


inline float3 operator+(const float3 &p1,const float3 &p2)
{
  return float3(
    p1.x+p2.x,
    p1.y+p2.y,
    p1.z+p2.z
  );
}


inline float3 operator^(const float3 &p1, const float3 &p2)
{
  return float3(
    +p1.y*p2.z-p2.y*p1.z,
    -p1.x*p2.z+p2.x*p1.z,
    +p1.x*p2.y-p2.x*p1.y
  );
}


inline float operator*(const float3 &p1, const float3 &p2)
{
  return p1.x*p2.x+p1.y*p2.y+p1.z*p2.z;
}


inline float3 operator*(const float f, const float3 &p2)
{
  return float3(f*p2.x, f*p2.y, f*p2.z);        
}


inline float3 operator*(const float3 &p2, const float f)
{
  return f*p2;
}

//inline float operator*(const float3 &p1, const float3 &p2)

inline float3 &operator*=(float3 &p1, float a)
{
  p1.x*=a;
  p1.y*=a;
  p1.z*=a;
  return p1;
}


//-----------------------------------------------------------------------------

Matrix::Matrix (float diag_element)
{
  for (int i=0; i<4;i++)
    for (int j=0; j<4;j++)
      m[i][j]=0.0;
  for (int i=0; i<4;i++)
    m[i][i]=diag_element;
}


Matrix &Matrix::Invert()
{
  double d;
  Matrix Result;
  d=+ m[0][0]*m[1][1]*m[2][2]
    + m[1][0]*m[2][1]*m[0][2]
    + m[2][0]*m[0][1]*m[1][2]
    - m[2][0]*m[1][1]*m[0][2]
    - m[0][0]*m[2][1]*m[1][2]
    - m[1][0]*m[0][1]*m[2][2];
  Result.m[0][0]= (m[1][1]*m[2][2] - m[2][1]*m[1][2]) / d;
  Result.m[0][1]= (m[2][1]*m[0][2] - m[0][1]*m[2][2]) / d;
  Result.m[0][2]= (m[0][1]*m[1][2] - m[1][1]*m[0][2]) / d;
  Result.m[0][3]= -(+ m[0][3]*m[1][1]*m[2][2]
              + m[1][3]*m[2][1]*m[0][2]
              + m[2][3]*m[0][1]*m[1][2]
              - m[2][3]*m[1][1]*m[0][2]
              - m[0][3]*m[2][1]*m[1][2]
              - m[1][3]*m[0][1]*m[2][2]) / d;
  Result.m[1][0]= (m[2][0]*m[1][2] - m[1][0]*m[2][2]) / d;
  Result.m[1][1]= (m[0][0]*m[2][2] - m[2][0]*m[0][2]) / d;
  Result.m[1][2]= (m[1][0]*m[0][2] - m[0][0]*m[1][2]) / d;
  Result.m[1][3]=-(+ m[0][0]*m[1][3]*m[2][2]
             + m[1][0]*m[2][3]*m[0][2]
             + m[2][0]*m[0][3]*m[1][2]
             - m[2][0]*m[1][3]*m[0][2]
             - m[0][0]*m[2][3]*m[1][2]
             - m[1][0]*m[0][3]*m[2][2]) / d;
  Result.m[2][0]= (m[1][0]*m[2][1] - m[2][0]*m[1][1]) / d;
  Result.m[2][1]= (m[2][0]*m[0][1] - m[0][0]*m[2][1]) / d;
  Result.m[2][2]= (m[0][0]*m[1][1] - m[1][0]*m[0][1]) / d;
  Result.m[2][3]=-(+ m[0][0]*m[1][1]*m[2][3]
             + m[1][0]*m[2][1]*m[0][3]
             + m[2][0]*m[0][1]*m[1][3]
             - m[2][0]*m[1][1]*m[0][3]
             - m[0][0]*m[2][1]*m[1][3]
             - m[1][0]*m[0][1]*m[2][3]) / d;
  Result.m[3][0]=0;
  Result.m[3][1]=0;
  Result.m[3][2]=0;
  Result.m[3][3]=1;
  return (*this=Result);
}


void Matrix::glLoad() const
{
  Matrix tmp_matrix;
  for (int i=0; i<4;i++)
    for (int j=0; j<4;j++)
      tmp_matrix.m[i][j]=m[j][i];
  glLoadMatrixf(tmp_matrix.m[0]);
}


void Matrix::glMult() const
{
  Matrix tmp;
  for (int i=0; i<4;i++)
    for (int j=0; j<4;j++)
      tmp.m[i][j]=m[j][i];
  glMultMatrixf(tmp.m[0]);
}


Matrix Matrix::Rotatex(float angle)
{
  using namespace std;
  Matrix new_matrix;
  int i,j;
  for (i=0; i<4;i++)
    for (j=0; j<4;j++)
      new_matrix.m[i][j]=0.0;
  new_matrix.m[0][0]=1;
  new_matrix.m[2][2]=(float)cos(angle);
  new_matrix.m[2][1]=+(float)sin(angle);
  new_matrix.m[1][2]=-(float)sin(angle);
  new_matrix.m[1][1]=(float)cos(angle);
  new_matrix.m[3][3]=1;
  return new_matrix;
}


Matrix Matrix::Rotatey(float angle)
{
  using namespace std;
  Matrix new_matrix;
  int i,j;
  for (i=0; i<4;i++)
    for (j=0; j<4;j++)
      new_matrix.m[i][j]=0.0;
  new_matrix.m[1][1]=1;
  new_matrix.m[0][0]=(float)cos(angle);
  new_matrix.m[0][2]=(float)sin(angle);
  new_matrix.m[2][0]=-(float)sin(angle);
  new_matrix.m[2][2]=(float)cos(angle);
  new_matrix.m[3][3]=1;
  return new_matrix;
}


Matrix Matrix::Rotatez(float angle)
{
  using namespace std;
  Matrix new_matrix;
  int i,j;
  for (i=0; i<4;i++)
    for (j=0; j<4;j++)
      new_matrix.m[i][j]=0;
  new_matrix.m[2][2]=1;
  new_matrix.m[0][0]=(float)cos(angle);
  new_matrix.m[0][1]=(float)sin(angle);
  new_matrix.m[1][0]=-(float)sin(angle);
  new_matrix.m[1][1]=(float)cos(angle);
  new_matrix.m[3][3]=1;
  return new_matrix;
}


Matrix Matrix::Translate(float delta_x, float delta_y, float delta_z)
{
  Matrix new_matrix;
  int i,j;
  for (i=0; i<4;i++)
    for (j=0; j<4;j++)
      new_matrix.m[i][j]=0;
  new_matrix.m[0][3]=delta_x;
  new_matrix.m[1][3]=delta_y;
  new_matrix.m[2][3]=delta_z;
  new_matrix.m[3][3]=
    new_matrix.m[2][2]=
    new_matrix.m[1][1]=
    new_matrix.m[0][0]=
    1;
  return new_matrix;
}

//&a != result, &b != &result
void MatrixMul(Matrix &result, const Matrix &a, const Matrix &b)
{
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++) {
      float s =
        a.m[i][0]*b.m[0][j]+
        a.m[i][1]*b.m[1][j]+
        a.m[i][2]*b.m[2][j]+
        a.m[i][3]*b.m[3][j];
      result.m[i][j]=s;
    }
}

Matrix operator* (const Matrix &a, const Matrix &b)
{
  Matrix new_matrix;
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++) {
      float s =
        a.m[i][0]*b.m[0][j]+
        a.m[i][1]*b.m[1][j]+
        a.m[i][2]*b.m[2][j]+
        a.m[i][3]*b.m[3][j];
      new_matrix.m[i][j]=s;
      /*float s=0;
      for (int k=0; k<4; k++)
        s+=a.m[i][k]*b.m[k][j];
      new_matrix.m[i][j]=s;*/
    }
  return new_matrix;
}


float3 operator* (const Matrix &m1, const float3 &m2)
{
  return float3(
    m1.m[0][0]*m2.x+m1.m[0][1]*m2.y+m1.m[0][2]*m2.z+m1.m[0][3],
    m1.m[1][0]*m2.x+m1.m[1][1]*m2.y+m1.m[1][2]*m2.z+m1.m[1][3],
    m1.m[2][0]*m2.x+m1.m[2][1]*m2.y+m1.m[2][2]*m2.z+m1.m[2][3]
  );
}


float3 operator&(const Matrix &m1, const float3 &m2)
{
  return float3(
    m1.m[0][0]*m2.x+m1.m[0][1]*m2.y+m1.m[0][2]*m2.z,
    m1.m[1][0]*m2.x+m1.m[1][1]*m2.y+m1.m[1][2]*m2.z,
    m1.m[2][0]*m2.x+m1.m[2][1]*m2.y+m1.m[2][2]*m2.z
  );
}



