// (c) Ivan Yanikov, dobrokot 'at' gmail.com

#ifndef MATIRIX_H_
#define MATIRIX_H_

#include <windows.h>
#include <cmath>
#pragma hdrstop

class Matrix
{
  public:
    float m[4][4];
    Matrix (float diag_element=1);
    void glMult() const;
    void glLoad() const;
    Matrix &Invert();
    static Matrix Rotatex(float angle);
    static Matrix Rotatey(float angle);
    static Matrix Rotatez(float angle);
    static Matrix Translate(float delta_x, float delta_y, float delta_z);
};


struct float3
{
  union {
    struct { float x,y,z; };
    float xyz[3];
  };
  float3(float fx, float fy, float fz) { x=fx; y=fy; z=fz; };
  float3() {};
  float3 &Normalize() {
    float t=1/std::sqrt(x*x+y*y+z*z);
    x*=t; y*=t; z*=t;
    return *this;
  }
};

float Length(const float3 &p);
inline float SqrLength (const float3 &p);
inline float3 &operator-=(float3 &p2, const float3 &p1);
inline float3 operator-(const float3 &p2, const float3 &p1);
inline float3 operator-(const float3 &p);
inline float3 &operator+=(float3 &p1,const float3 &p2);
inline float3 operator+(const float3 &p1,const float3 &p2);
inline float3 operator^(const float3 &p1, const float3 &p2);
inline float operator*(const float3 &p1, const float3 &p2);
inline float3 operator*(const float f, const float3 &p2);
inline float3 operator*(const float3 &p2, const float f);
inline float3 &operator*=(float3 &p1, float a);

//&a != result, &b != &result
void MatrixMul(Matrix &result, const Matrix &a, const Matrix &b);
Matrix operator* (const Matrix &a, const Matrix &b);
float3 operator* (const Matrix &m1, const float3 &m2);
float3 operator&(const Matrix &m1, const float3 &m2);

#endif
